
/* DYNAMIC PLAIN HUFFFMAN. -- 
A dynamic word-based byte oriented compressor for text files based on 
dynamic plain huffman.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <string.h>
#include <stdio.h>
#include "BufferedWriter.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



 /*------------------------------------------------------------------
 Creates a new buffer
 ------------------------------------------------------------------*/
BufferWrite bufCreateWriter (char *fileName)

  {BufferWrite B = (BufferWrite) malloc (sizeof(struct BWriter));

 	B->_fileName = (char * )malloc (strlen(fileName));
	strcpy(B->_fileName,fileName);

	B->size = WRITER_BUFFER_SIZE;
	B->data = (byte *) malloc (B->size);
	return B;
  }
/*------------------------------------------------------------------
 Flushes all the data in the buffer and frees memory
------------------------------------------------------------------*/
void destroyBW(BufferWrite bW)
{
	flushBW(bW);
	free (bW->data);
 }


//////////////////////////////////////////////////////////////////////
// PUBLIC METHODS
//////////////////////////////////////////////////////////////////////
/*------------------------------------------------------------------
 Opens file for writting
 ------------------------------------------------------------------*/
	void openBW(BufferWrite bW) {

		printf("Trying to open file: [%s] for writting \n",bW->_fileName);

//		if ((bW->outFile = fopen(bW->_fileName,"w")) == NULL)
		if ( (bW->outFile = open (bW->_fileName, O_WRONLY|O_CREAT, 0666)) ==-1)
		{
			//fprintf(stderr,"\nError while opening file: %s\n",bW->_fileName);
			//fprintf(stderr,"Program will exit now\n");
			return;
		}
		printf("File [%s] successfuly opened\n",bW->_fileName);

	}

/*------------------------------------------------------------------
 Closes the file
------------------------------------------------------------------*/
	void closeBW(BufferWrite bW) {
		flushBW(bW);
//		fclose(bW->outFile);
		close(bW->outFile);
	}

/*------------------------------------------------------------------
 Writes all the data of the buffer to disk, leaving the buffer empty
 ------------------------------------------------------------------*/
	void flushBW(BufferWrite bW) {
		if (bW->posInBuffer >0) {
			//			fwrite(bW->data, sizeof(byte), bW->posInBuffer, bW->outFile);
			write(bW->outFile,  bW->data,  bW->posInBuffer);

			bW->posInBuffer=0;
		}
	}



/*------------------------------------------------------------------
 Adds n bytes to the buffer bW starting from the pointer orig.
 If the buffer becomes full, the whole buffer is output to disk.
------------------------------------------------------------------*/
		long writeBW(BufferWrite bW, unsigned char *orig, long n) {
			byte *dst;
			register int i;
            //fprintf(stderr,"\n escribe [%s]",orig);
			if (bW->posInBuffer < WRITER_BUFFER_SIZE - n) {
					//	bcopy(orig, &(bW->data[bW->posInBuffer]) , n);
				    dst = (byte *) bW->data + bW->posInBuffer;
 			        for (i=1;i<=n;i++) {*(dst++) = *(orig++); }
					bW->posInBuffer += n;
			}
			else {  //hay que volcar el buffer a disco (hasta posInBuffer)
					write(bW->outFile,  bW->data,  bW->posInBuffer);
				    //y copiar la palabra al comienzo del nuevo buffer;
					dst = (byte *) bW->data;
 			        for (i=1;i<=n;i++) {*(dst++) = *(orig++); }
					bW->posInBuffer = n;
			}
			//flushBW(bW);
		}



/*------------------------------------------------------------------
 Adds n bytes to the buffer bW starting from the pointer orig. It
 also adds a '\0' character after those "n" bytes.
 If the buffer becomes full, the whole buffer is output to disk.
------------------------------------------------------------------*/

		long writeBWPlusZero(BufferWrite bW, unsigned char *orig, long n) {
			byte *dst;
			register int i;
            //fprintf(stderr,"\n escribe [%s]",orig);
			if (bW->posInBuffer < WRITER_BUFFER_SIZE - (n+1)) {
					//	bcopy(orig, &(bW->data[bW->posInBuffer]) , n);
				    dst = (byte *) bW->data + bW->posInBuffer;
 			        for (i=1;i<=n;i++) {*(dst++) = *(orig++); }
 			        (*dst) = '\0';
					 bW->posInBuffer += n+1;
			}
			else {  //hay que volcar el buffer a disco (hasta posInBuffer)
					write(bW->outFile,  bW->data,  bW->posInBuffer);
				    //y copiar la palabra al comienzo del nuevo buffer;
					dst = (byte *) bW->data;
 			        for (i=1;i<=n;i++) {*(dst++) = *(orig++); }
 			        (*dst) = '\0'; 			        
					bW->posInBuffer = n+1;
			}
			//flushBW(bW);
		}
