
/* DYNAMIC PLAIN HUFFMAN. -- 
A dynamic word-based byte oriented compressor for text files based on 
dynamic plain-Huffman.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*-----------------------------------------------------------------------
 File      : compressor
 Function  : Main file of the compressor
 ------------------------------------------------------------------------*/


#include "compressor.h"
#include "version.h"



/*------------------------------------------------------------------
 Initilizes the structures used.
 ---------------------------------------------------------------- */
void initialize (unsigned long tamArq){
	unsigned long i;

    initialize_hash(tamArq);

	N_int = N/B_value +1;

	_memMgr = createMemoryManager(); //_memMgr.initialize();


	intNodes   = (t_IntNode*) malloc (N_int * sizeof(t_IntNode));
	nodeIndex  = (t_NodeIndex*) malloc ((N + N_int+10) * sizeof(t_NodeIndex));

	zeroNode = 1;
	lastIntNode = 0;

	intNodes[lastIntNode].minChild = 1;
	intNodes[lastIntNode].maxChild = 1;
	intNodes[lastIntNode].nodePos  = 0;

	//initialization of the root of the tree
	nodeIndex[0].relPos  = lastIntNode;
	nodeIndex[0].nodeType= IS_INTERNAL;
	nodeIndex[0].parent = 999999999;

	//Initialization of the zeroNode
	nodeIndex[zeroNode].nodeType= IS_LEAF;
	nodeIndex[zeroNode].parent  = 0;

	//there is a list of Blocks

	nodeInBlock	= (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));
	nextBlock	= (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));
	previousBlock = (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));
	topBlock	= (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));
	lastBlock	= (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));
	weight		= (unsigned long *) malloc ((N+ N_int)*sizeof(unsigned long));

	availableBlock	= 1;

	nodeInBlock[0]	      = 0;
	nodeInBlock[zeroNode] = 0;
	previousBlock[0]      = 0;
	nextBlock[0]	      = 0;
	weight[0]	          = 0;

	topBlock[0] 	      = 0;
	lastBlock[0]	      = 1;

	for (i=availableBlock ; i<=N+ N_int-2 ; i++) nextBlock[i]=i+1;
	nextBlock[N-1] = N+ N_int; //there is no nextBlock

}



/*------------------------------------------------------------------
 * InterchangeNodesLL
 * Interchanges two leaf nodes with the same frequency
 * i,j is the position of the leaves in the nodeIndex
------------------------------------------------------------------ */
void interchangeNodesLL(unsigned long i, unsigned long j) {
	unsigned long tmp;

	hash[nodeIndex[i].relPos].nodePos = j;
	hash[nodeIndex[j].relPos].nodePos = i;

	tmp = nodeIndex[i].relPos;
	nodeIndex[i].relPos = nodeIndex[j].relPos;
	nodeIndex[j].relPos = tmp;


}


/*------------------------------------------------------------------
 * InterchangeNodesIL
 * Interchanges two nodes: a leaf node and an internal node
 *									nodes have the same frequency
 * leaf,internal is the position of both nodes in the nodeIndex
------------------------------------------------------------------ */
void interchangeNodesIL(unsigned long leaf, unsigned long internal) {
	unsigned long tmp;

	hash[nodeIndex[leaf].relPos].nodePos = internal;
	intNodes[nodeIndex[internal].relPos].nodePos = leaf;

	tmp = nodeIndex[leaf].relPos;
	nodeIndex[leaf].relPos = nodeIndex[internal].relPos;
	nodeIndex[internal].relPos = tmp;

	tmp = nodeIndex[leaf].nodeType;
	nodeIndex[leaf].nodeType = nodeIndex[internal].nodeType;
	nodeIndex[internal].nodeType = tmp;

	
}


/*------------------------------------------------------------------
 * InterchangeNodesIi
 * Interchanges two internal nodes:
 *									nodes have the same frequency
------------------------------------------------------------------ */
void interchangeNodesII(unsigned long I, unsigned long J) {
	unsigned long tmp;
	
	intNodes[nodeIndex[I].relPos].nodePos = J;
	intNodes[nodeIndex[J].relPos].nodePos = I;

	tmp = nodeIndex[I].relPos;
	nodeIndex[I].relPos = nodeIndex[J].relPos;
	nodeIndex[J].relPos = tmp;

	
}



/*------------------------------------------------------------------
 * GetCode
 * gets the codeword of a word ranked "position"
 * returns an unsigned long value (up to 4 bytes codewords) in "*codeword"
 *         the number of bytes of the codeword.
------------------------------------------------------------------ */
void getCode(register unsigned long position) {
	register unsigned long q, pq;
	register unsigned long codeword, kk;

	q=position;
	pq = nodeIndex[position].parent;


	codeword = (position - intNodes[pq].minChild);
	q = intNodes[pq].nodePos;
	kk = 1;


	while (q != 0) {

		pq = nodeIndex[q].parent;
	
		codeword += (q - intNodes[pq].minChild) * BASE[kk];
		q = intNodes[pq].nodePos;
		kk++;
	}
	
	CODEWORD =codeword;
	K=kk;
}

/*------------------------------------------------------------------
 Obtains the size of the file to compress.
------------------------------------------------------------------ */
unsigned long fileSize (char *filename){
	FILE *fpText;
	unsigned long fsize;
	fpText = fopen(filename,"rb");
	fsize=0;
	if (fpText) {
		fseek(fpText,0,2);
		fsize= ftell(fpText);
		fclose(fpText);
		fprintf(stderr,"fileSize = %ld",fsize);
	}
	return fsize;
}

/*------------------------------------------------------------------
 * GetCodeAndWrite
------------------------------------------------------------------ */
void getCodeAndWrite(BufferWrite bw,register unsigned long position) {
	register unsigned long q, pq;
	register unsigned long kk;
	
	
	unsigned char *codeword;
	
	codeword =(unsigned char *) (code+8);

	q=position;
	pq = nodeIndex[position].parent;

	(*codeword) = (position - intNodes[pq].minChild);
	q = intNodes[pq].nodePos;
	kk = 1;

	while (q != 0) {
		pq = nodeIndex[q].parent;

		(*(--codeword)) = (q - intNodes[pq].minChild);		
		q = intNodes[pq].nodePos;
		kk++;
	}
		
	K=kk;
	


	writeBW(bw,codeword,kk);
}



/*------------------------------------------------------------------
 * FindNode(j)
 * sets q as the node that should be slided&incremented in next step
 * if j was not previously in the vocabulary (j was the zeroNode)
 *  then "word j" is added to the vocabulary.
 *  freq(j) is not increased yet.
------------------------------------------------------------------ */
void findNode() {
	unsigned long pq,npq;
	unsigned long bq,nbq,bqI;


	q=j;
	bq=nodeInBlock[q];

	if (q ==zeroNode) {
		pq = nodeIndex[q].parent;
		if ((intNodes[pq].maxChild - intNodes[pq].minChild ) < B_value -1) {
			//pq had less than B_value children, so there is enough space to hold "q"

				//Handle blocks list
				nbq = nextBlock[bq];

				if (weight[nbq]== 1) {  //block "1" already exists
					nodeInBlock[q] = 1;
					nodeInBlock[q+1] = 0;

					lastBlock[1] = q;
					lastBlock[0] = q+1;
					topBlock[0] = q+1;
				}
				else {    //block "1" did not exists

					bqI = availableBlock;
					availableBlock = nextBlock[availableBlock];

					nextBlock[bqI]= nextBlock[bq];
					previousBlock[bqI]=previousBlock[nbq];
					nextBlock[bq] = bqI;
					previousBlock[nbq]= bqI;

					weight[bqI] = 1;

					nodeInBlock[q] = 1;
					nodeInBlock[q+1] = 0;

					topBlock[bqI]=q;
					lastBlock[bqI]=q;
					topBlock[bq]=q+1;
					lastBlock[bq]=q+1;

				}


				//q is added to the tree
			 	insertElement ((unsigned char *) CURRENTWORD, SIZEWORD, &addrInTH);
			 	hash[addrInTH].nodePos = q;
				nodeIndex[q].relPos = addrInTH;
				nodeIndex[q].nodeType = IS_LEAF;
				nodeIndex[q].parent = pq;


				//zeroNode is increased
				zeroNode++;
				
				nodeIndex[zeroNode].nodeType = IS_LEAF;
				nodeIndex[zeroNode].parent = pq;

				//q's parent child counter is also increased
				intNodes[pq].maxChild ++;

		}
		else {//a new internal node has to be created
				//the new internal node is created
				lastIntNode++;
				npq = lastIntNode;
				nodeIndex[q].relPos = npq;
				nodeIndex[q].nodeType = IS_INTERNAL;
				nodeIndex[q].parent = pq;
				intNodes[npq].minChild = q+1;
				intNodes[npq].maxChild = q+2;
				intNodes[npq].nodePos = q;


				//Handle blocks list
				nbq = nextBlock[bq];

				if (weight[nbq]== 1) {  //block "1" already exists
					nodeInBlock[q] = 1;
					nodeInBlock[q+1] =1;
					nodeInBlock[q+2] = 0;

					lastBlock[1] = q+1;
					lastBlock[0] = q+2;
					topBlock[0] = q+2;
				}
				else {    //block "1" did not exists

					bqI = availableBlock;
					availableBlock = nextBlock[availableBlock];

					nextBlock[bqI]= nextBlock[bq];
					previousBlock[bqI]=previousBlock[nbq];
					nextBlock[bq] = bqI;
					previousBlock[nbq]= bqI;

					weight[bqI] = 1;

					nodeInBlock[q] = 1;
					nodeInBlock[q+1] = 1;
					nodeInBlock[q+2] = 0;

					topBlock[bqI]=q;
					lastBlock[bqI]=q+1;
					topBlock[bq]=q+2;
					lastBlock[bq]=q+2;

				}

				// The new word is added to the tree.
				q++;
				insertElement ((unsigned char *) CURRENTWORD, SIZEWORD, &addrInTH);
				hash[addrInTH].nodePos = q;
				nodeIndex[q].relPos = addrInTH;
				nodeIndex[q].nodeType = IS_LEAF;
				nodeIndex[q].parent = npq;

				//zeroNode is updated
				zeroNode+=2;
				//nodeIndex[zeroNode].relPos = -1;
				nodeIndex[zeroNode].nodeType = IS_LEAF;
				nodeIndex[zeroNode].parent = npq;
		}
		//the new node to slide&increment is now pq;
		q = intNodes[pq].nodePos;

	}
	else { //q was not the zeroNode
			 //so q was already the node to be slided and increased
	}
	
}




/*------------------------------------------------------------------
 * IncreaseBlock
 * Increases the weight of "q", changing "q" of block if needed.
------------------------------------------------------------------ */
void increaseBlock() {



				unsigned long b, bq, nbq;

				bq = nodeInBlock[q];
				nbq = nextBlock[bq];

				if ( weight[nbq] == (weight [bq] +1) ) {

					nodeInBlock[q] = nbq;
					lastBlock[nbq] = q;

					if (lastBlock [bq] == q) { //q's old block disappears since q was the only word in block bq;
						previousBlock[nbq] = previousBlock[bq];
						nextBlock[previousBlock[bq]] = nbq;
						nextBlock[bq] = availableBlock;
						availableBlock =bq;
					}
					else {
						topBlock[bq] = q+1;
					}

				}
				else{ //Next block did not exist.
					if (lastBlock[bq] == q) {
						//it was the only element in its group
						//works fine also for zero-node
						weight[bq] = weight[bq] +1;
					}
					else {   //there were more nodes in its group
						//create a new block (nbq, bq) since block of weight = weight[bq] +1
						//did not exist.

						b = availableBlock;
						availableBlock = nextBlock[availableBlock];
						nextBlock[bq] = b;
						previousBlock[b] = bq;
						nextBlock[b] = nbq;
						previousBlock[nbq] = b;

						topBlock[bq] = q+1;
						topBlock[b] = q;
						lastBlock[b] = q;
						weight[b] = weight[bq] +1;
						nodeInBlock[q] = b;
					}
				}


   q = intNodes[nodeIndex[q].parent].nodePos;

}


/*------------------------------------------------------------------
 * Update.
 * Structures are updated. Two steps are needed.
 *		-FindNode(j): Puts word "j" as the top of its block.
 *		-IncreaseBlock: Changes "j" of block (to a block of weight +1)
------------------------------------------------------------------ */
void update() {
	unsigned long top;

	findNode(); //sets q as the node to slide and increment

	while ( q != 0 ) { //0 is the root of the tree...
		//top = getTop(q);
		top= topBlock[nodeInBlock[q]];

		if (q!=top) {
			if (nodeIndex[q].nodeType == IS_LEAF ) {  //q is a leaf node
				if (nodeIndex[top].nodeType == IS_LEAF ) {
					interchangeNodesLL (q,top);

				}
				else {
					interchangeNodesIL (q,top);

				}

				q=top;
			}
			else { //q is an internal node
				if (nodeIndex[top].nodeType == IS_LEAF ) {
					interchangeNodesIL (top,q);

				}
				else {


					interchangeNodesII (top,q);
				}
				q=top;
			}
		}

		//q is the new top of its block
		increaseBlock();
	}
	//root is also increased
	weight[0]++;


}




/*------------------------------------------------------------------
 * Performs the full Encoding process.
------------------------------------------------------------------ */
void mainCompressor (char* infile, char* outfile, unsigned long N_val) {



   BufferRead B;
	unsigned char *text,*top, *nextbuf;
	register unsigned char *pbeg,*pend,*wordstart;
	register unsigned long size;
	BufferWrite bw;

	register unsigned long i;

	int f;

	/**/

   StartValid();



	f = open (infile, O_RDONLY);
	if (f == -1)
		printf("Cannot read file %s",infile);
	B = bufCreate(MAX_SIZE_OF_WORD+5);    

	bufSetFile (B,f);



	bw= bufCreateWriter(outfile);
	openBW(bw);


  initialize(N_val); //initializes compressor for up to N_value words
  fprintf(stderr, "\n\n\nSTARTING COMPRESSION...[%ld]", B->fpos);






	if (bufEmpty(B) ) {size =0; return;}
	//ELSE BYTES REMAINS TO BE READ... AND FIRST IS AN ALREADY READ BYTE
	size = 0;

	while ( B->size==B->dsize ) {

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
        //fprintf(stderr,"\n Entra 1 pbeg [%d] pend [%d] valid [%d]",pbeg,pend, _Valid[*pbeg]);
	 	while ( pbeg < pend )   {

			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
	
			while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] )) {
		
						size++;
						pbeg++;
					}

		    }
	      else {
			if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
            
				while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])) {

					size++;
					pbeg++;
				}
			}
			else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not


	                pbeg++;
					if (_Valid [*pbeg] ) {
						wordstart = pbeg;
						while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}

					}
					else {   // a "separator word" is parsed until an alphanumerical character is found

						size++;
						while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )) {
							size++;
							pbeg++;
						}
					}//else {  // a "separator word"
			}

	  }



  /*********************************/
  /*********************************/
		
		CURRENTWORD=wordstart;
		SIZEWORD = size;

      j = search ((unsigned char *)CURRENTWORD, SIZEWORD, &addrInTH );
	 	if (j==WORD_NOT_FOUND) j=zeroNode;

		
	  	getCodeAndWrite(bw,j);
		

		
		 if (j==zeroNode) {
			// if currentWord is a new word, then the "currentWord" is also sent
			// notice that the '\0' caracter in "currentWord" is also sent as mark to
			// enable decoder to distinguish where that word ends.
			writeBWPlusZero(bw,(unsigned char *)CURRENTWORD,SIZEWORD);
		 }
   	 	/**/

		update();

  /*********************************/
  /*********************************/

    	}/* while interno*/

    	nextbuf=pbeg;
      bufLoad (B, (char*)nextbuf);

	}//while(1)









if (!bufEmpty(B)){
	 	bufCurrent (B,(char**)&text,(char**)&top);
	 	pbeg = text; pend = top;

		while ( pbeg < pend )   {

			size=0;
			wordstart = pbeg;
			if (_Valid[*pbeg]) {   //alphanumerical data
		    	while ( (size<MAX_SIZE_OF_WORD) && (  _Valid[*pbeg] ) &&( pbeg < pend )) {

						size++;
						pbeg++;
					}
		    }
	      else {
				if (*pbeg != ' ') { //a separator comes starting in ' ' comes, so it is a new word
	
					while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg])&&( pbeg < pend )) {
	
						size++;
						pbeg++;
					}
				}
				else {  //a  SPACE comes, so we have to test if next character is alphanumerical or not
	
		                pbeg++;
						if (_Valid [*pbeg] ) {
							wordstart = pbeg;
							while ( (size<MAX_SIZE_OF_WORD) && (_Valid[*pbeg] )&&( pbeg < pend )) {
								size++;
								pbeg++;
							}
	
						}
						else {   // a "separator word" is parsed until an alphanumerical character is found
	
							size++;
							while ( (size<MAX_SIZE_OF_WORD) && (!_Valid[*pbeg] )&&( pbeg < pend )) {
								size++;
								pbeg++;
							}
						}//else {  // a "separator word"
	
				}

			  }



  /*********************************/
  /*********************************/

		CURRENTWORD=wordstart;
		SIZEWORD = size;

      j = search ((unsigned char *)CURRENTWORD, SIZEWORD, &addrInTH );
	 	if (j==WORD_NOT_FOUND) j=zeroNode;

	
	  	getCodeAndWrite(bw,j);
		

		
		 if (j==zeroNode) {
			// if currentWord is a new word, then the "currentWord" is also sent
			// notice that the '\0' caracter in "currentWord" is also sent as mark to
			// enable decoder to distinguish where that word ends.
			writeBWPlusZero(bw,(unsigned char *)CURRENTWORD,SIZEWORD);
		 }
   	 	/**/

		update();

   /*********************************/
   /*********************************/



    	}/* while interno*/
   }//if (!bufEmpty(B){




	bufDestroy(B);
	close(f);
	closeBW(bw);

	free(hash);


} //mainCompressor



/*------------------------------------------------------------------ */
/*
 *  COMPRESSOR PROGRAM OF DYNAMIC DPH
 */
/*------------------------------------------------------------------ */

int main(int argc, char* argv[])
{

	char fileName[20];
	long sizeNValue;
	unsigned long bytesFile;

	showMessage();
	
	sizeNValue=MAX_LEN_VOCABULARY;
	
	
	
	printf("ARGC VALE [%d]\n",argc);
	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to DPH dynamic compressor\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> <[VOC SIZE]>\n",argv[0]);
		return 0;
	}

	//Estimation (Using Heap's law) of the size of the vocabulary (in number of distinct words).
	
	bytesFile = fileSize(argv[1]);
	if (!bytesFile) {
		fprintf(stderr, "\nFILE EMPTY OR FILE NOT FOUND !!\nCompression aborted\n");
		exit(0);
	}

	if(bytesFile<5000000) bytesFile = 5000000;
	sizeNValue = (unsigned long) floor(3.9* pow(bytesFile,0.69) );
	
	if (argc ==4) 
		  sizeNValue = atol(argv[3]);
		

	fprintf(stderr,"\nMax size of voc set[%ld]",sizeNValue);


	StartRusageTime();
	
	
	mainCompressor(argv[1], argv[2],sizeNValue);
	//StopRusageTime();
	StopRusageTimeFile("timesCDPH",argv[1]);
	


}


