
/* DYNAMIC PLAIN HUFFMAN. -- 
A dynamic word-based byte oriented compressor for text files based on 
dynamic plain-Huffman.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <malloc.h>
#include <stdio.h>

//#include "hash.h"
#include "MemoryManager.h"
#include "constants.h"

#include "BufferedReader.h"
#include "writer.h"         //#include "../IO/BufferedWriter.h"
#include "timer.h"

#define IS_LEAF 0			   //node type regards to a leaf
#define IS_INTERNAL 1	   //node type regards to an internal node
#define B_value 256





	MemoryManager _memMgr;


	//information of words
	struct Nword {
		unsigned char *word;
		unsigned int size;
		unsigned long nodePos;

	};
	typedef struct Nword t_word;
	
	struct NIntNodes {
		  unsigned long minChild;	/* first child of the node  (remember that they are adyancent in the NodeIndex) */
		  unsigned long maxChild;	/* last  child of the node */
		  unsigned long nodePos;	/* position occupied in the nodeIndex. */
	};
	
	typedef struct NIntNodes t_IntNode;


	struct NNodeIndex {
		  unsigned long relPos;		/* relative position to the hash table (leaves) or */
		  									/* to the intNodes array */
		  unsigned long nodeType;	/* type of node: 0 = leaf , 1 = internal node */
		  unsigned long parent;		/* position in the intNodes array of the parent of this node */		  
	};
	
	typedef struct NNodeIndex t_NodeIndex;


 	/* private attributes */

	//variables used in the "full process"
	unsigned long j;	 	           //stores the position in the vocabulary of the current parsed word
	unsigned long q;  			 	   /* Node of the current word  */

	unsigned long N;				   /* max number of words that can be held */
	unsigned long N_int;			   /* max number of internal nodes */


   //variables used in DECODING process
	char *CURRENTWORD;	//parsed word
	unsigned int SIZEWORD;   		 		    //size of parsed word



	//Array that holds the data that regards to the internal nodes of the tree
	t_IntNode *intNodes;		
	unsigned long lastIntNode;		 /* last position used in internal nodes structure  */

	//Array that holds the data that regards to words: leaf nodes of the tree
	t_word   *words;     		 	 /* leaf nodes of the tree */
	unsigned long lastLeafNode;

	//Array that holds the NodeIndex that holds all nodes of the tree
	t_NodeIndex *nodeIndex;		

	//information of blocks
	unsigned long *nodeInBlock;		 /* v[i] = x , word i is in block x */
	unsigned long *nextBlock;         /* v[i] = x , next block of i is x */
	unsigned long *previousBlock;	 /* v[i] = x , prev block of i is x */
	unsigned long *topBlock;			 /* v[i] = x , if word x is the top of block i */
	unsigned long *lastBlock;		 /* v[i] = x , if word x is the last word of block i */
	unsigned long *weight;			 /* v[i] = x , if block i is weighted x */
	
	//0-node and availableBlock
	unsigned long zeroNode;			 /* zeroNode is the place where a new word will be inserted */
	unsigned long availableBlock;     /* Place where to find an available block, to create a new one */

	unsigned long K;


	Writer outWriter;


	void initialize (unsigned long tamArq);
	void insertElement (const unsigned char *aWord, register unsigned long len , register unsigned long position);

	void mainDecompressor (char* infile, char* outfile, unsigned long N_val);


